package com.roidapp.imagelib.facedetector.orionvision

import android.content.Context
import com.orion.vision.bean.Face
import com.roidapp.imagelib.facedetector.FaceDetectorInterface
import comroidapp.baselib.util.Logger
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel

open class OrionVisionFaceDetector(private val context: Context,
                              private var width: Int,
                              private var height: Int,
                              private val callback: FaceDetectorInterface.FaceDetectorCallback<Face>) : FaceDetectorInterface {

    private var faceDetectorJob: Job? = null
    private val faceTracker: OrionVisionFaceTracker by lazy {
        OrionVisionFaceTracker(context)
    }
    private var faceDetectorLatch : Channel<Boolean>? = null

    init {
        init()
    }

    override fun start() {
        synchronized(this) {
            if (Logger.ENABLE_LOG) Logger.d("start")

            faceDetectorJob = GlobalScope.launch(Dispatchers.Default) { doFaceDetectionProcedure(this) }
        }
    }

    @ExperimentalCoroutinesApi
    override fun resume() {
        GlobalScope.launch(Dispatchers.Default) {
            if (Logger.ENABLE_LOG) Logger.d("resuming face detector...")
            faceDetectorLatch?.let { latch ->
                if (latch.isEmpty) latch.send(true)
            }
        }
    }

    override fun pause() {
        faceDetectorLatch = when(faceDetectorLatch) {
            is Channel<Boolean> -> faceDetectorLatch
            else -> {
                if (Logger.ENABLE_LOG) Logger.d("pausing face detector...")
                Channel()
            }
        }
    }

    override fun isPaused(): Boolean {
        return faceDetectorLatch != null
    }

    override fun stop() = runBlocking {
        synchronized(this) {
            launch {
                if (Logger.ENABLE_LOG) Logger.d("stopping face detector...")
                faceDetectorJob?.cancelAndJoin()
                if (Logger.ENABLE_LOG) Logger.d("face detector stop success!")
                callback.onStopDetection()
            }
        }
    }

    override fun setViewSize(_width: Int, _height: Int) {
        width = _width
        height = _height
        faceTracker.clearTrack()
    }

    private fun init() {
        GlobalScope.launch(Dispatchers.Default) {
            if (!faceTracker.isLoaded()) {
                Logger.w("Failed to init OrionVision native lib!!!")
                return@launch
            }

            withContext(Dispatchers.Main) {
                callback.onLibraryLoaded()
            }
        }
    }

    // Trying to figure out the param among com.orion.vision.bean.Parameter, which represents proper direction of the camera.
    private fun getDirectionParam(direction: Int): Int {
        // Since PG will tackle front/rear camera difference, we always config detector in rear camera case.
        // The role of 90 and 270 degrees rotation must be switched.
        return if (direction and 1 != 0) (direction xor 2) else direction
    }

    override suspend fun doFaceDetectionProcedure(scope: CoroutineScope) {

        withContext(Dispatchers.Main) { callback.onStartDetection() }

        while (scope.isActive) {

            faceDetectorLatch?.let { latch ->
                withContext(Dispatchers.Main) {
                    callback.onPauseDetection()
                }
                latch.receive() // coroutine will suspend here until something sent at the other side of the channel
                faceDetectorLatch = null
                withContext(Dispatchers.Main) {
                    callback.onResumeDetection()
                }
            }

            callback.getCameraPixelArray()?.let { cameraNV21ByteArray ->
                val pixelArray = ByteArray(cameraNV21ByteArray.size)

                synchronized(cameraNV21ByteArray) {
                    if (pixelArray.size >= cameraNV21ByteArray.size) {
                        System.arraycopy(cameraNV21ByteArray, 0, pixelArray, 0, cameraNV21ByteArray.size)
                    }
                }

                if (Logger.ENABLE_LOG) {
                    Logger.d("pixelArray size = ${pixelArray.size}, nvArray size = ${cameraNV21ByteArray.size}, " +
                            "width = $width, height = $height, " +
                            "dir = ${callback.getDirection()}, rotation = ${callback.getRotation()}, " +
                            "camera orientation = ${callback.getCameraOrientation()}, " +
                            "device orientation = ${callback.getDeviceOrientation()}, " +
                            "dir param = ${getDirectionParam(callback.getDirection())}")
                }
                callback.onFaceDataUpdated(faceTracker.track(pixelArray, width, height, getDirectionParam(callback.getDirection())))
            }
        }

        // Comes to here to close the tracker.
        if (Logger.ENABLE_LOG) Logger.d("closing face tracker...")
        faceTracker.close()
        if (Logger.ENABLE_LOG) Logger.d("face tracker closed")
    }
}