package com.roidapp.imagelib.facedetector.orionvision

import android.content.Context
import android.graphics.Bitmap
import com.orion.vision.OrionVision
import com.orion.vision.bean.Face
import com.roidapp.imagelib.facedetector.FaceTrackerInterface
import comroidapp.baselib.util.Logger

import java.io.BufferedReader

class OrionVisionFaceTracker(private val context: Context) : FaceTrackerInterface<Face> {

    companion object {
        private const val ORION_VISION_LICENSE_FILE_NAME = "orion_vision_license.txt"
        private const val ORION_VISION_LIBRARY_NAME = "orionvision-lib"

        private const val FACE_RECOGNITION_FACTOR = -1.0f
        private const val FACE_QUALITY_FACTOR = 0.3f
        private const val FACE_KEYPOINT_FACTOR = 0.3f
        private const val FACE_TEST_FACTOR = -1.0f
    }

    private var nativeLoaded: Boolean = false

    private val orionVisionLicense: String by lazy {
        context.resources.assets.open(ORION_VISION_LICENSE_FILE_NAME).bufferedReader().use(BufferedReader::readText)
    }

    init {
        try {
            System.loadLibrary(ORION_VISION_LIBRARY_NAME)
        } catch (e: UnsatisfiedLinkError) {}

        init()
    }

    override fun init(): Boolean {
        if (!configOrionVision() || !initOrionVision()) {
            Logger.w("Failed to init OrionVision native lib!!!")
            return false
        }

        if (Logger.ENABLE_LOG) Logger.w("Loading OrionVisionFaceTracker successful")
        nativeLoaded = true
        return true
    }

    // return false if license test failed. Otherwise get true
    private fun configOrionVision(): Boolean {
        return when(OrionVision.getInst().license(context, orionVisionLicense)) {
            true -> {
                OrionVision.getInst().setModels(OrionVision.Models.FACE_RECOGNITION, true)
                OrionVision.getInst().setModels(OrionVision.Models.FACE_RECOGNITION, FACE_RECOGNITION_FACTOR)
                OrionVision.getInst().setModels(OrionVision.Models.FACE_KEYPOINT, true)
                OrionVision.getInst().setModels(OrionVision.Models.FACE_KEYPOINT, FACE_KEYPOINT_FACTOR)
                OrionVision.getInst().setModels(OrionVision.Models.EXTRACT_FACE_FEATURE, false)
                true
            }
            false -> false
        }
    }

    private fun initOrionVision(): Boolean {
        return OrionVision.getInst().initModel() == 0
    }

    override fun isLoaded(): Boolean {
        synchronized(this) {
            return nativeLoaded
        }
    }

    override fun track(bytes: ByteArray, width: Int, height: Int, direction: Int): ArrayList<Face> {
        synchronized(this) {
            return if (nativeLoaded) {
                OrionVision.getInst().facekeypoint(bytes,  width, height, direction)
            } else {
                ArrayList()
            }
        }
    }

    override fun track(bitmap: Bitmap): List<Face> {
        synchronized(this) {
            return if (nativeLoaded) {
                OrionVision.getInst().facekeypoint(bitmap)
            } else {
                ArrayList()
            }
        }
    }

    override fun clearTrack() {
        synchronized(this) {
            if (nativeLoaded) OrionVision.getInst().cleartracker()
        }
    }

    override fun close() {
        synchronized(this) {
            nativeLoaded = false
            OrionVision.getInst().fkpdestroy()
        }
    }
}